#define ANI_TIME .1

#define MAX_PART 1000
#define MAX_BULLET 1000
#define MAX_OBJECT 100
#define MAX_SPEED 1000

#define LEVEL_CLEAR 0
#define LEVEL_LOST 8

#define FALSE 0
#define TRUE -1
#define NULL -1

#define BULLET_GOOD 1
#define BULLET_BAD -1

#include "fmod.bi"

enum
    EFFECT_NONE
    EFFECT_SHIELDS
    EFFECT_WEAPON
    EFFECT_BLOCK
    EFFECT_LIFE
    EFFECT_UFO
    EFFECT_UFO_WEAPON
end enum

enum 
    BULLET_NORM
    BULLET_INV
end enum

enum
    C_ESC = 1
    C_DEV =15
    C_PAUSE = 25
    C_SHOOT = 57
    C_LEFT = 75
    C_RIGHT = 77    
end enum

type Vec_t
    as integer x,y
end type

type Img_t
    as integer ptr img 'img data
    as integer ptr ptr spr 'points to sprites
    as integer w,h
    as short cFrame
    as short maxFrames
    as single lu    
end type

type Part_t
    as Vec_t p,v
    as integer s
    union
        as single lu,ll
    end union
    
end type

type Bullet_t
    as Vec_t p,v
    as integer s,o,kind'sprite,damage,owner,type of bullet
    union
        as single lu
        as integer dmg
    end union
end type

type Weapon_t
    as integer dmg,v,s,bS,eBs 'damage,velocity,sprite,bullet sprite,enemy bullet sprite
    as single r 'reload time
    as string n 'name
end type

type Enemy_t
    as Vec_t p
    
    union
        shields as integer
        minX as integer 
    end union 'health, grid Y size
    
    union 
        s  as integer 
        maxX as integer 
    end union 'sprite, grid X size   
    
    union 
        w as Weapon_t 
        n(1) as integer 
    end union 'weapon, number of enemies/direction
    
    ls as single  'last shot
    as integer wN 'weapon number
end type

type Player_t
    as integer x,xS,s,y,sS'shield sprite
    as integer shields,lives
    as single lu
    as single ls,sLH'shields last hit
    as integer w 'weapon index number
end type

type Object_t
    as Vec_t p,v
    as integer fx,s,fxAm 'effect,sprite, effects amount
    as single lu
end type

type Star_t
    union
        as Vec_t p
        as single lu
    end union
    
    as Vec_t v
    
    as integer col    
end type

type Engine_t
    as Vec_t cam,scr
    as byte cPg
    as integer level,fps,tFps,eLft,tSa 'level, fps,temp fps, starfield y, enemies left
    as single tm,eOl
    as byte fScr
    
    as Img_t ptr i
    as Part_t ptr part
    as Bullet_t ptr b
    as Player_t ptr play
    as Enemy_t ptr e
    as Object_t ptr o
    as Weapon_t ptr w
    as Star_t ptr s
end type

'*******************************************************************************
'Art Subs
'*******************************************************************************
sub ImgLoad (i as Img_t ptr, imgFile as string, frames as ubyte)
    dim as uinteger bmpWidth,bmpHeight,bmpSize,mem, cFrame
    FF = freefile
    
    open imgFile+"0.bmp" for binary as #FF
        get #FF, 19, bmpWidth
        get #FF, 23, bmpHeight
    close #FF
    
    bmpSize=bmpHeight*bmpWidth+1
    
    if bmpSize=1 then exit sub

    with *i
        .img=callocate(bmpSize*(frames+1)*sizeof(.img))
        .spr=callocate(        (frames+1)*sizeof(.spr))
        
        .w=bmpWidth
        .h=bmpHeight
        
        for cFrame = 0 to frames
            
            bload imgfile+str$(cFrame)+".bmp", @.img[bmpSize*cFrame]
            
            .spr[cFrame]=@.img[bmpSize*cFrame]
            
        next
        
        .maxFrames=frames    
        .cFrame =0
    end with
            
end sub

sub LoadArt (eng as Engine_t ptr)
    dim as integer tC, cI
    dim as string$ dat,nam
    
    FF = freefile
    
    open "data\sprites.txt" for input as #FF

        do
            input #FF, dat
        loop while left$(dat,2) = "//"

        eng->i=callocate((val(dat)+1)*sizeof(Img_t))
    
        eng->i[0].maxFrames=val(dat)
        
        for tC= 1 to val(dat)
            
            do
                input #FF, dat
            loop while left$(dat,2)="//" 'ignore comments
            
            nam=dat
            
            do
                input #FF, dat
            loop while left$(dat,2)="//"'ignore comments
            
            ImgLoad(@eng->i[tC],"art\"+nam,val(dat)-1)
            
            'set initial animation data
            with eng->i[tC] 
                .cFrame=0
                .maxFrames=val(dat)-1
                .lu=timer
            end with
            
        next
        
    close #FF
    
end sub

sub UpdateArt(eng as Engine_t ptr)
    dim as integer tC
    
    for tC = 1 to eng->i[0].maxFrames
        with eng->i[tC]
            if (timer - .lu > ANI_TIME) then
                .cFrame+=1
                if (.cFrame > .maxFrames) then .cFrame=0
                .lu=timer
            end if
        end with
    next
end sub

sub PauseArt(eng as Engine_t ptr)
    dim as integer tC
    
    for tC = 1 to eng->i[0].maxFrames
        eng->i[tC].lu=timer-eng->i[tC].lu
    next
    
end sub


sub DrawArt(eng as Engine_t ptr, spr as integer, x as integer, y as integer)
    
    if spr=NULL then exit sub
    
    with eng->i[spr]
        put(x,y),.spr[.cFrame],trans
    end with
    
end sub

sub DrawArtA(eng as Engine_t ptr, spr as integer, x as integer, y as integer, a as integer)
    
    if spr=NULL then exit sub
    if a > 255 then a = 255
    if a < 0 then a = 0
    
    with eng->i[spr]
        put(x,y),.spr[.cFrame],alpha,a
    end with
    
end sub

sub KillArt(eng as Engine_t ptr)
    deallocate(eng->i)
end sub

'*******************************************************************************
'Particle Subs
'*******************************************************************************
sub InitPart(eng as Engine_t ptr)
    eng->part=callocate(sizeof(Part_t))
    eng->part[0].s=0
end sub

sub SpawnPart(eng as Engine_t ptr, num as integer, x as integer, y as integer, spr as integer)
    dim as integer tC,tS,tA
    
    eng->part->lu=timer
    
    for tC = 1 to eng->part[0].s
        with eng->part[tC]
            if .ll <= 0 then
                num-=1
                .p.x=x*10
                .p.y=y*10
                
                tS=rnd*150+120
                tA=rnd*6282
                
                .v.x=cos(tA/1000)*tS*3
                .v.y=sin(tA/1000)*tS*3
                
                .ll=rnd*.5
                
                .s=spr
            end if
        end with
        
        if num <= 0 then exit for
    
    next
    
    if num > 0 then
        eng->part[0].s+=num
        if eng->part[0].s > MAX_PART then eng->part[0].s=MAX_PART
        
        eng->part=reallocate(eng->part,(eng->part[0].s+1)*sizeof(Part_t))
        
        for tC = eng->part[0].s-num+1 to eng->part[0].s
            
            with eng->part[tC]
                
                .p.x=x*10
                .p.y=y*10
                
                tS=rnd*150+120
                tA=rnd*6282
                
                .v.x=cos(tA/1000)*tS*3
                .v.y=sin(tA/1000)*tS*3
                
                .ll=rnd*.5
                
                .s=spr            
            end with
        
        next        
    end if

end sub

sub UpdatePart(eng as Engine_t ptr)
    dim as integer tC
    dim as single lu
    
    lu=timer-eng->part->lu
    
    if lu < .05 then exit sub
    
    do
        tC+=1
        if tC > eng->part[0].s then exit do
        
        with eng->part[tC]
            
            if .ll > 0 then
                .p.x+=.v.x*lu
                .p.y+=.v.y*lu
                .ll-=lu
            else
                
                .p.x=eng->part[eng->part[0].s].p.x
                .p.y=eng->part[eng->part[0].s].p.y
                
                .v.x=eng->part[eng->part[0].s].v.x
                .v.y=eng->part[eng->part[0].s].v.y
                
                .ll=eng->part[eng->part[0].s].ll
                
                .s=eng->part[eng->part[0].s].s
                
                eng->part=reallocate(eng->part,eng->part[0].s*sizeof(Part_t))
                
                eng->part[0].s-=1
                
                tC-=1
                
            end if
        
        end with
    loop
    
    eng->part->lu=timer
    
end sub

sub PausePart(eng as Engine_t ptr)
    dim as integer tC
    
    eng->part->lu=timer-eng->part->lu
    
end sub

sub DrawPart(eng as Engine_t ptr)
    dim as integer tC,a
    
    for tC = 1 to eng->part[0].s
        
        with eng->part[tC]
            a=.ll*510'255
            if .ll > 0 then DrawArtA(eng,.s,.p.x/10,.p.y/10,a)
        end with
    
    next
    
end sub

sub KillPart(eng as Engine_t ptr)
    deallocate(eng->part)
end sub
'*******************************************************************************
'Collision
'*******************************************************************************
function Collide (x1 as integer,y1 as integer,w1 as integer,h1 as integer,x2 as integer,y2 as integer,w2 as integer,h2 as integer) as byte
    if x1 < x2+w2-1 then
        if x1+w1-1 > x2 then
            if y1 < y2+h2-1 then
                if y1+h1-1 > y2 then
                    return TRUE
                end if
            end if
        end if
    end if
    return FALSE
end function

'*******************************************************************************
'Weapons
'*******************************************************************************
'type Weapon_t
'    as integer dmg,v,s,bS 'damage,velocity,sprite,bullet sprite
'    as single r 'reload time
'    as string n 'name
'end type

sub InitWeapon(eng as Engine_t ptr)
    dim as integer FF,num,tC
    dim as string dat    
    
    FF = freefile
    open "data\weapon.txt" for input as #FF
        do
            input #FF,dat
        loop while left$(dat,2)="//"
        
        num = val(dat)
        
        eng->w=callocate(sizeof(weapon_t)*(num+1))
        eng->w[0].s=num
        
        for tC = 1 to eng->w[0].s
            
            with eng->w[tC]
            
                do
                    input #FF,dat
                loop while left$(dat,2)="//"
                
                .n=dat
                
                do
                    input #FF,dat
                loop while left$(dat,2)="//"
                
                .s=val(dat)
                
                do
                    input #FF,dat
                loop while left$(dat,2)="//"
                
                .dmg=val(dat)
                
                do
                    input #FF,dat
                loop while left$(dat,2)="//"
                
                .bS=val(dat)
                
                do
                    input #FF,dat
                loop while left$(dat,2)="//"
                
                .eBS=val(dat)
                
                do
                    input #FF,dat
                loop while left$(dat,2)="//"
                
                .r=val(dat)
                
                do
                    input #FF,dat
                loop while left$(dat,2)="//"
                
                .v=val(dat)

            end with
            
        next
    
    close #FF
end sub

sub KillWeapon(eng as Engine_t ptr)
    deallocate(eng->w)
end sub
           
'*******************************************************************************
'Bullets
'*******************************************************************************
sub InitBullet(eng as Engine_t ptr)
    eng->b=callocate(sizeof(Bullet_t))
    eng->b[0].s=0
end sub

sub SpawnBullet(eng as Engine_t ptr, x as integer, y as integer,xS as integer, yS as integer,dmg as integer, spr as integer, o as integer)
    dim as integer tC,tS,tA,num
    
    num=1
    eng->b->lu=timer
    for tC = 1 to eng->b[0].s
        with eng->b[tC]
            if .dmg <= 0 then
                
                num-=1
                .p.x=x*10
                .p.y=y*10
                
                .v.x=xS*10
                .v.y=yS*10
                
                .dmg=dmg
                
                .s=spr
                .o=o
                .kind=kind
            end if
        end with
        
        if num <= 0 then exit for
    
    next
    
    if num > 0 then
        eng->b[0].s+=num
        if eng->b[0].s > MAX_BULLET then eng->b[0].s=MAX_BULLET
        
        eng->b=reallocate(eng->b,(eng->b[0].s+1)*sizeof(Bullet_t))
        
        for tC = eng->b[0].s-num+1 to eng->b[0].s
            
            with eng->b[tC]
                
                .p.x=x*10
                .p.y=y*10
                
                .v.x=xS*10
                .v.y=yS*10
                
                .dmg=dmg
                
                .lu=timer              
                
                .s=spr
                .o=o
                .kind=kind
            end with
        
        next        
    end if
    
end sub

sub UpdateBullet(eng as Engine_t ptr)
    dim as integer tC
    dim as single lu
    
    lu=timer-eng->b->lu
    
    if lu < .05 then exit sub
    
    for tC = 1 to eng->b[0].s
        
        with eng->b[tC]
            
            if .dmg > 0 then
                .p.x+=.v.x*lu
                .p.y+=.v.y*lu
                
                if .p.y < 0 then .dmg=0
                if .p.y > 2400 then .dmg=0
                
            end if            
        
        end with
    next
    
    eng->b->lu=timer
    
end sub

sub PauseBullet(eng as Engine_t ptr)
    dim as integer tC
    
    eng->b->lu=timer-eng->b->lu
    
end sub

sub DrawBullet(eng as Engine_t ptr)
    dim as integer tC
    
    for tC = 1 to eng->b[0].s
        
        with eng->b[tC]
            if .dmg > 0 then DrawArt(eng,.s,.p.x/10,.p.y/10)
        end with
    
    next
    
end sub

sub KillBullet(eng as Engine_t ptr)
    deallocate(eng->b)
end sub

'*******************************************************************************
'Player
'*******************************************************************************
sub InitPlayer(eng as Engine_t ptr,x as integer, y as integer,spr as integer, w as integer)
    eng->play=callocate(sizeof(Player_t))
    with *eng->play
        .s=spr
        .x=x*10
        .y=y*10
        .lu=timer
        .ls=timer-10
        .shields=3
        .lives=5
        .sS=36
        .sLH=timer-2
        .w=w

    end with
end sub

sub UpdatePlayer(eng as Engine_t ptr)
    
    with *eng->play
        if .lu=0 then
            .lu=timer
            exit sub
        end if
        
        
        if timer-.lu > .05 then
            .lu=timer-.lu
            
            '.x+=.xS
            if .shields < 1 then
                .lives-=1
                .shields=3
                SpawnPart(eng,30,.x/10,.y/10,3)
                .x=(eng->scr.x/2-8)*10
            end if
            
            if multikey(c_Left) then
                'if .xS > 0 then .xS=0
                .x-=MAX_SPEED*.lu
            elseif multikey(c_Right) then
                'if .xS < 0 then .xS=0
                .x+=MAX_SPEED*.lu
            else
                .xS=0'*=(1-.lu)
            end if
            
            if .x < 0 then 
                .x=0
                .xS=0
            elseif .x > (eng->scr.x-eng->i[.s].w)*10 then
                .x=(eng->scr.x-eng->i[.s].w)*10
                .xS=0
            end if
            
            if abs(.xS) > MAX_SPEED then .xS = MAX_SPEED * sgn(.xS)
            
            .lu=timer
        
        end if
        
        if multikey(c_Shoot) then
            if timer-.ls >= eng->w[.w].r then
                
                SpawnBullet(eng,.x/10+eng->i[.s].w shr 1-2,.y/10,0,-abs(eng->w[.w].v),eng->w[.w].dmg,eng->w[.w].bS,BULLET_GOOD)
                
                .ls=timer
                
            end if
        end if
        
        
    end with
end sub

sub PausePlayer(eng as Engine_t ptr)
    eng->play->lu=timer-eng->play->lu
    eng->play->ls=timer-eng->play->ls
end sub

sub DrawPlayer(eng as Engine_t ptr)
    with *eng->play
        DrawArt(eng,.s,.x/10,.y/10)
        if timer-.sLH < .15 then DrawArtA(eng,.sS,.x/10,.y/10,255-1700*(timer-.sLH))        
    end with
end sub

sub KillPlayer(eng as Engine_t ptr)
    deallocate(eng->play)
end sub

'*******************************************************************************
'Enemy
'*******************************************************************************
'type Enemy_t
'    as Vec_t p
'    union: hp as integer: minX as integer :end union 'health, grid Y size
'    union: s  as integer: maxX as integer :end union 'sprite, grid X size   
'    union: w as Weapon_t: num as integer  :end union 'weapon, number of enemies
'    ls as single  'last shot
'end type

sub InitEnemy(eng as Engine_t ptr,mFile as string)
    dim as string dat
    dim as integer r,c,tE,tR,tC
    
    FF=freefile
    open mFile for input as #FF
        do
            input #FF, dat
        loop while left$(dat,2)="//"
        
        c=val(dat)
        
        do
            input #FF, dat
        loop while left$(dat,2)="//"
        
        r=val(dat)
        
        eng->e=callocate((c*r+1)*sizeof(Enemy_t))
        
        with eng->e[0]
            .n(0)=c*r
            .n(1)=1
            .ls=timer
            .p.x=(eng->scr.x/2-c*12)*10
            .p.y=120
            .minX=1
            .maxX=c
        end with
        
        tE=1
        
        for tR = 1 to r
            for tC = 1 to c
                
                with eng->e[tE]
                    do
                        input #FF, dat
                    loop while left$(dat,2)="//"
                
                
                    .p.x=tC*240-240
                    .p.y=tR*160-160
                    .s=val(dat)
                    
                    do
                        input #FF, dat
                    loop while left$(dat,2)="//"
                    
                    .w.dmg=eng->w[val(dat)].dmg
                    .shields=val(dat)
                    .w.v=eng->w[val(dat)].v
                    .w.s=eng->w[val(dat)].s
                    .w.bS=eng->w[val(dat)].eBS
                    .w.r=eng->w[val(dat)].r*3+3
                    .wN=val(dat)
                    .ls=timer-rnd*6
                    
                end with
                
                tE+=1
            next
        next
    
    close #FF

end sub

sub UpdateEnemy(eng as Engine_t ptr)
    dim as integer tC,tSa,spr
    
    with eng->e[0]
        
        .ls=timer-.ls
        
        if .ls > .05 then
            if eng->eLft < 40 then tSa=160
            if eng->eLft < 20 then tSa=320
            if eng->eLft < 10 then tSa=640
            if eng->eLft < 5 then tSa=960
            if eng->eLft = 3 then tSa=1200
            if eng->eLft = 2 then tSa=1400
            if eng->eLft = 1 then tSa=2400
            
            .p.x+=(80+tSa)*.ls*.n(1)
            
            if .p.x+.maxX*240-90 > eng->scr.x*10 then 
                .p.x=eng->scr.x*10-.maxX*240+90
                .p.y+=60
                .n(1)=-1
            elseif .p.x+.minX*240-240 < 0 then
                .p.x=240-.minX*240
                .p.y+=60
                .n(1)=1
            end if
            
            .ls=timer
            
        else
            .ls=timer-.ls
        end if
    
    end with
    
    for tC=1 to eng->e[0].n(0)
        
        if eng->e[tC].shields > 0 then
            
            with eng->e[tC]
            
                if timer-.ls > .w.r then
                    
                    if (int(rnd*10) = 4) then
                        'enemy fires a bullet
                        if .w.dmg = 1 then spr=8
                        if .w.dmg = 2 then spr=14
                        if .w.dmg = 3 then spr=15
                        SpawnBullet(eng,(.p.x+eng->e[0].p.x)/10,(.p.y+eng->e[0].p.y)/10,0,50,.w.dmg,spr,BULLET_BAD)
                    end if
                    
                    .ls=timer-rnd
                    
                end if
                
            end with
            
        end if
        
    next    
        
end sub

sub PauseEnemy(eng as Engine_t ptr)
    dim as integer tC
    
    
    eng->e[0].ls=timer-eng->e[0].ls
    
    for tC=1 to eng->e[0].n(0)
        eng->e[tC].ls=timer-eng->e[tC].ls
    next    
        
end sub

sub CheckEnemy(eng as Engine_t ptr)
    dim as integer tC,cMin=10000000,cMax=-100,yMax
    
    eng->eLft=0
    
    for tC = 1 to eng->e[0].n(0)
        
        with eng->e[tC]
            
            if .shields > 0 then
                if .p.x < cMin then cMin=.p.x
                if .p.x > cMax then cMax=.p.x
                
                if .p.y+eng->e[0].p.y > yMax then yMax=.p.y+eng->e[0].p.y
            
                eng->eLft+=1
            end if
            
        
        end with
        
    next
    
    if eng->eLft > 0 then
        eng->e[0].minX=cMin/240+1
        eng->e[0].maxX=cMax/240+1
        
        'if enemies are at player level then game over
        if yMax+160 > eng->play->y then eng->e[0].n(1)=LEVEL_LOST
        
    else
        eng->e[0].n(1)=LEVEL_CLEAR
    end if
    
end sub

sub DrawEnemy(eng as Engine_t ptr)
    dim as integer tC
    
    for tC = 1 to eng->e[0].n(0)
        
        with eng->e[tC]
            if .shields > 0 then DrawArt(eng,.s,(.p.x+eng->e[0].p.x)/10,(.p.y+eng->e[0].p.y)/10)
        end with
    
    next
    
end sub

sub KillEnemy(eng as Engine_t ptr)
    deallocate(eng->e)
end sub

'*******************************************************************************
'Object
'*******************************************************************************

'type Object_t
'    as Vec_t p,v
'    as integer fx,s 'effect,sprite
'    as single lu
'end type

sub InitObject(eng as Engine_t ptr)
    eng->o=callocate(sizeof(Object_t))
    eng->o[0].s=0
end sub

sub SpawnObject(eng as Engine_t ptr, x as integer, y as integer,xS as integer, yS as integer,fx as integer,fxAm as integer, spr as integer)
    dim as integer tC,tS,tA,num
    
    num=1
    
    for tC = 1 to eng->o[0].s
        with eng->o[tC]
            if .fx = EFFECT_NONE then
                num-=1
                .p.x=x*10
                .p.y=y*10
                
                .v.x=xS*10
                .v.y=yS*10
                
                .fx=fx
                .fxAm=fxAm
                
                .lu=timer
                
                .s=spr
            end if
        end with
        
        if num <= 0 then exit for
    
    next
    
    if num > 0 then
        eng->o[0].s+=num
        if eng->o[0].s > MAX_OBJECT then eng->o[0].s=MAX_OBJECT
        
        eng->o=reallocate(eng->o,(eng->o[0].s+1)*sizeof(Object_t))
        
        for tC = eng->o[0].s-num+1 to eng->o[0].s
            
            with eng->o[tC]
        
                num-=1
                .p.x=x*10
                .p.y=y*10
                
                .v.x=xS*10
                .v.y=yS*10
                
                .fx=fx
                .fxAm=fxAm
                
                .lu=timer
                
                .s=spr
            end with
        
        next        
    end if

end sub

sub UpdateObject(eng as Engine_t ptr)
    dim as integer tC
    with *eng
        for tC = 1 to .o[0].s
            
            with .o[tC]
                
                if .fx <> EFFECT_NONE then
                    .lu=timer-.lu
                    
                    if .lu >= .05 then
                    
                        .p.x+=.v.x*.lu
                        .p.y+=.v.y*.lu
                        
                        if .p.x > eng->scr.x*10 then
                            .fx=EFFECT_NONE
                        end if
                        
                        if .p.x+eng->i[.s].w*10 < 0 then
                            .fx=EFFECT_NONE
                        end if                    
                        
                        .lu=timer
                    else
                        .lu=timer-.lu
                    end if
                
                end if            
            
            end with
        next
    end with
    
end sub

sub PauseObject(eng as Engine_t ptr)
    dim as integer tC
    
    for tC = 1 to eng->o[0].s
        eng->o[tC].lu=timer-eng->o[tC].lu
    next
    
end sub

sub DrawObject(eng as Engine_t ptr)
    dim as integer tC
    
    with *eng
        for tC = 1 to .o[0].s
            
            with .o[tC]
                if .fx <> EFFECT_NONE then DrawArt(eng,.s,.p.x/10,.p.y/10)
            end with
        next
    end with
end sub

sub KillObject(eng as Engine_t ptr)
    deallocate(eng->o)
end sub

'*******************************************************************************
'Star Subs
'*******************************************************************************

sub InitStars(eng as Engine_t ptr, num as integer)
    dim as integer tC,tCol,tR
    
    eng->s=callocate(sizeof(Star_t)*(num+1))
    
    eng->s->col=num
    eng->s->lu=timer
    for tC=1 to num
        with eng->s[tC]
            .p.x=int(rnd*320)
            .p.y=int(rnd*720)
            .v.x=0
            tR=int(rnd*40)
            '.v.y=-(tR*tR)/40+40
            .v.y=int(rnd*40)+20
            .col=rgb((.v.y)*4,(.v.y)*4,(.v.y)*4)            
        end with
    next
    
end sub

sub UpdateStars(eng as Engine_t ptr)
    dim as integer tC
    dim as single lu
    
    lu=timer-eng->s->lu
    
    if lu < .05 then exit sub
    
    for tC=1 to eng->s->col
        
        with eng->s[tC]
            
            if timer-.lu > .05 then
                
                .p.y+=.v.y*lu
                
                if .p.y > 720 then .p.y-=720
                
            end if
            
        end with
        
    next
    
    eng->s->lu=timer
    
end sub

sub PauseStars(eng as Engine_t ptr)
    dim as integer tC
    
    eng->s->lu=timer-eng->s->lu

    
end sub

sub DrawStars(eng as Engine_t ptr)
    dim as integer tC
    
    for tC = 1 to eng->s->col
        
        with eng->s[tC]
            pset(.p.x,.p.y),.col
        end with
        
    next
    
end sub

sub KillStars(eng as Engine_t ptr)
    deallocate(eng->s)
end sub

'*******************************************************************************
'Object interaction
'*******************************************************************************
'player-bullet,enemy-bullet,object-bullet,player-object,enemy-object

sub CheckInteraction(eng as Engine_t ptr)
    dim as integer tB, tE,tObj
    
    'bullet stuff first
    with *eng
    
        for tB=1 to .b[0].s
            if .b[tB].dmg > 0 then
            
                if .b[tB].o=BULLET_BAD then
                    if Collide(.b[tB].p.x,.b[tB].p.y,.i[.b[tB].s].w*10,.i[.b[tB].s].h*10, .play->x,.play->y,.i[.play->s].w*10,.i[.play->s].h*10) then
                        
                        'bullet hits player
                        .play->shields-=.b[tB].dmg
                        
                        if .play->shields > 0 then
                            .play->sLH=timer
                        end if
                        
                        .b[tB].dmg=0
                        
                    end if
                elseif .b[tB].o=BULLET_GOOD then
                    
                    for tE = 1 to .e[0].n(0)
                        
                        if .e[tE].shields > 0 then
                            
                            if Collide(.b[tB].p.x,.b[tB].p.y,.i[.b[tB].s].w*10,.i[.b[tB].s].h*10,(.e[tE].p.x+.e[0].p.x),(.e[tE].p.y+.e[0].p.y),.i[.e[tE].s].w*10,.i[.e[tE].s].h*10) then
                                
                                'bullet hits enemy
                                .e[tE].shields-=.b[tB].dmg
                                if .e[tE].shields <= 0 then 
                                    SpawnPart(eng,40,(.e[tE].p.x+.e[0].p.x)/10,(.e[tE].p.y+.e[0].p.y)/10,3)
                                    if .e[tE].wN > .play->w then
                                        if int(rnd*10)=1 then
                                            SpawnObject(eng,(.e[tE].p.x+.e[0].p.x)/10,(.e[tE].p.y+.e[0].p.y)/10,0,25,EFFECT_WEAPON,.e[tE].wN,.e[tE].w.s)
                                        end if
                                    end if
                                    
                                end if
                                
                                if not(.b[tB].kind AND BULLET_INV) then .b[tB].dmg=0
                                
                            end if
                        end if
                    next
                end if
            
                'see if bullets collide with blocking object
                for tObj = 1 to .o[0].s
                    
                    if (.o[tObj].fx = EFFECT_BLOCK) or (.o[tObj].fx = EFFECT_UFO) then
                        if Collide(.b[tB].p.x,.b[tB].p.y,.i[.b[tB].s].w*10,.i[.b[tB].s].h*10,  .o[tObj].p.x,.o[tObj].p.y,.i[.o[tObj].s].w*10,.i[.o[tObj].s].h*10) then
                            
                            with .o[tObj]
                                if (.fx = EFFECT_UFO) then
                                    
                                    'SpawnObject(o as Object_t ptr, x,y,xSpeed,ySpeed,effect,effect Amount, sprite)
                                    SpawnObject(eng,.p.x/10,.p.y/10,0,100,EFFECT_LIFE,1,5)
                                    .fx=EFFECT_NONE
                                end if
                                
                                eng->b[tB].dmg=0
                            
                            end with
                        
                        end if
                    end if
                
                next
                
            end if
        next    
        
        'object-player/enemy
        for tObj = 1 to .o[0].s
            
            'enemy
            if .o[tObj].fx = EFFECT_BLOCK then
                
                for tE = 1 to .e[0].n(0)
                    
                    if .e[tE].shields > 0 then
    
                        if Collide(.o[tObj].p.x,.o[tObj].p.y,.i[.o[tObj].s].w*10,.i[.o[tObj].s].h*10,    .e[tE].p.x+.e[0].p.x,.e[tE].p.y+.e[0].p.y,.i[.e[tE].s].w*10,.i[.e[tE].s].h*10) then
                            
                            'enemy hits blocking object
                            .o[tObj].fxAm-=.e[tE].shields
                            if .o[tObj].fxAm <= 0 then .o[tObj].fx=EFFECT_NONE
                            
                            .e[tE].shields=0
                            SpawnPart(eng,20,(.e[tE].p.x+.e[0].p.x)/10,(.e[tE].p.y+.e[0].p.y)/10,3)
                            
                        
                        end if
                    end if
                next
            
            'object with player
            elseif .o[tObj].fx <> EFFECT_NONE then
                
                if Collide(.o[tObj].p.x,.o[tObj].p.y,.i[.o[tObj].s].w*10,.i[.o[tObj].s].h*10,  .play->x,.play->y,.i[.play->s].w*10,.i[.play->s].h*10) then
        
                    if .o[tObj].fx     =EFFECT_HEATH then
                        .play->shields+=.o[tObj].fxAm
                    elseif .o[tObj].fx =EFFECT_LIFE then
                        .play->lives+=.o[tObj].fxAm   
                    elseif .o[tObj].fx=EFFECT_WEAPON then
                        .play->w=.o[tObj].fxAm
                    end if
                    
                    .o[tObj].fx=EFFECT_NONE
                    
                end if
            end if
            
        next
    
    end with
end sub
'*******************************************************************************
'Font
'*******************************************************************************
sub CPrint(txt as string,x as integer,y as integer,eng as Engine_t ptr)
    dim as integer tC
    
    with *eng
        
        for tC = 0 to len(txt)-1
            
            if (txt[tC] > 96 and txt[tC] < 123) then
                put (x+8*tC,y),eng->i[9].spr[txt[tC]-97],trans
                
            elseif (txt[tC] > 64 and txt[tC] < 91) then
                put (x+8*tC,y),eng->i[9].spr[txt[tC]-65],trans
                
            elseif (txt[tC] > 47 and txt[tC] < 59) then
                put (x+8*tC,y),eng->i[10].spr[txt[tC]-48],trans        
            end if
            
        next
        
    end with
    
end sub

'*******************************************************************************
'Game Routines
'*******************************************************************************
sub CleanUp(eng as Engine_t ptr)
    
    KillPart(eng)
    KillBullet(eng)
    KillPlayer(eng)
    KillEnemy(eng)
    KillObject(eng)
    KillStars(eng)
    
end sub

sub PauseGame(eng as Engine_t ptr)
    
    PauseArt(eng)
    PausePart(eng)
    PauseEnemy(eng)
    PauseBullet(eng)
    PauseObject(eng)
    PausePlayer(eng)
    PauseStars(eng)
    eng->tm=timer-eng->tm
    
    screenset -(eng->cPg-1),-(eng->cPg-1)
    CPrint ("Paused",eng->scr.x/2-28,eng->scr.y/2-4,eng)
    
    'wait for p to be pressed again
    do:loop while multikey(C_PAUSE)
    do
        sleep 10
    loop until (multikey(C_PAUSE) or multikey(C_ESC))
    do:loop while multikey(C_PAUSE)
    
    PauseArt(eng)
    PausePart(eng)
    PauseEnemy(eng)
    PauseBullet(eng)
    PauseObject(eng)
    PausePlayer(eng)
    PauseStars(eng)
    eng->tm=timer-eng->tm    

end sub

sub BeginLevel(eng as Engine_t ptr)
    dim as single tTm
    dim as integer tC, tM
    
    eng->level+=1
    if eng->level < 6 then 
        if eng->level > 1 then
            eng->eOl=0
            KillPart(eng)
            KillBullet(eng)
            KillEnemy(eng)
            KillObject(eng)
            KillStars(eng)
        end if
        
        pS=6
        
        eng->eLft=100
        eng->eOl=0
        screenset 0,0
        cls    
        CPrint ("Level "+str$(eng->level),eng->scr.x/2-32,eng->scr.y/2-4,eng)
        
        do
            tM=0
            for tC = 1 to 127
                if multikey(tC) then tM+=1
            next
        loop until tM=0
        
        do
            for tC = 1 to 127
                if multikey(tC) then exit do
            next
        loop
        
        InitPart(eng)
        
        InitWeapon(eng)
        InitBullet(eng)
        if eng->level = 1 then InitPlayer(eng,eng->scr.x/2-8,eng->scr.y-eng->i[pS].h-4,pS,1)
        
        InitEnemy(eng,"data\m"+str$(eng->level-1)+".txt")
        
        InitObject(eng)
        InitStars(eng,150)
        
        SpawnObject(eng,20,210,0,0,EFFECT_BLOCK,2,4)
        SpawnObject(eng,148,210,0,0,EFFECT_BLOCK,2,4)
        SpawnObject(eng,276,210,0,0,EFFECT_BLOCK,2,4)
    end if    
    
    
end sub
'*******************************************************************************
'Engine
'*******************************************************************************
sub SetScreen(eng as Engine_t ptr)

    screenres eng->scr.x,eng->scr.y,32,2,eng->fScr
    
end sub

sub InitGame(eng as Engine_t ptr)
    dim as integer tC,tCol
    randomize timer
    
    eng->cPg=0
    
    eng->scr.x=320
    eng->scr.y=240
    
    SetScreen(eng)
    
    LoadArt(eng)
    
end sub

    
sub KillGame(eng as Engine_t ptr)
    KillArt(eng)
end sub

sub ShowEnd(eng as Engine_t ptr)
    dim as integer cY
    
    eng->tm=timer
    cY=eng->scr.y
    
    do
    
        cls
        
        CPrint ("AS THE FINAL RUINS GO FLYING OFF INTO",eng->scr.x/2-148,cY,eng)
        CPrint ("SPACE YOU SURVEY THE DAMAGE YOU HAVE",eng->scr.x/2-144,cY+12,eng)
        CPrint ("WRECKED UPON THE ALIEN HORDES",eng->scr.x/2-116,cY+24,eng)
        
        CPrint ("EARTH IS SAVED THANKS TO YOUR VALIANT",eng->scr.x/2-148,cY+48,eng)
        CPrint ("EFFORTS MANKIND OWES ITS EXISTANCE TO",eng->scr.x/2-148,cY+60,eng)
        CPrint ("YOU",eng->scr.x/2-12,cY+72,eng)
        
        CPrint ("AS YOU RELAX IN THE COCKPIT OF YOUR",eng->scr.x/2-140,cY+96,eng)
        CPrint ("SHUTTLE A LITTLE RED LIGHT GOES OFF",eng->scr.x/2-140,cY+108,eng)
        CPrint ("IT LOOKS LIKE THE BATTLE ISNT OVER YET",eng->scr.x/2-152,cY+120,eng)
        
        if timer-eng->tm > .1 then
            cY-=10*(timer-eng->tm)
            eng->tm=timer
        end if

        screenset eng->cPg, -(eng->cPg-1)
        eng->cPg=(eng->cPg-1) 
            
        if multikey(C_ESC) then exit do
        
    loop while (cY > -128)

end sub

'*******************************************************************************
'Main
'*******************************************************************************

sub Main(eng as Engine_t ptr)
    
    dim as integer pS,tC,tG,tD,tOb
    
    eng->cPg=0
    
    BeginLevel(eng)
    
    eng->tm=timer+1
    
    eng->eOl=0
    
    do
        tC=0
        for bla=0 to 127
            if multikey(bla) then tC+=1
        next
        if tC=0 then exit do
    loop
    
    do
        
        UpdateObject(eng)
        UpdateBullet(eng)
        UpdatePart(eng)
        UpdatePlayer(eng)
        UpdateEnemy(eng)
        CheckEnemy(eng)
        UpdateArt(eng)
        UpdateStars(eng)
        
        CheckInteraction(eng)
        
        line(0,0)-(319,239),0,bf 'erase the level
        
        if eng->e[0].n(1)=LEVEL_CLEAR then 
            
            if eng->eOl=0 then eng->eOl=timer
            
            CPrint ("Level Cleared",eng->scr.x/2-52,eng->scr.y/2-4,eng)
            
            if (timer-eng->eOl)>3 then BeginLevel(eng)
            if eng->level=6 then 
                ShowEnd(eng)
                exit sub
            end if
            
                        
        end if
        
        DrawStars(eng)
        DrawPart(eng)
        DrawEnemy(eng)
        DrawBullet(eng)
        DrawObject(eng)
        DrawPlayer(eng)
        
        
        if multikey(C_DEV) then 
            CPrint ("Fps:"+str$(eng->fps),0,0,eng)
            CPrint ("Parts:"+str$(eng->part->s),0,8,eng)
            CPrint ("Bullets:"+str$(eng->b->s),0,16,eng)
            CPrint ("Objects:"+str$(eng->o->s),0,24,eng)
        end if
        
        CPrint ("Lives: "+str$(eng->play->lives),0,eng->scr.y-9,eng)
        if eng->e[0].n(1) = LEVEL_LOST OR eng->play-> lives < 0 then
            screenset -(eng->cPg-1),-(eng->cPg-1)
            CPrint ("Game Over",128,116,eng)
            sleep 4000,1
            exit do
        end if
        
        
        screenset eng->cPg,-(eng->cPg-1)
        eng->cPg=-(eng->cPg-1)
        
        eng->tfps+=1
        if eng->tm < timer then
            eng->fps=eng->tFps
            eng->tFps=0
            eng->tm=timer+1
            
            if int(rnd*8)=1 then 
                tD=int(rnd*2)
                SpawnObject(eng,(eng->scr.x-1-eng->i[12].w)*tD,0,-100*(tD*2-1),0,EFFECT_UFO,EFFECT_SHIELDS,12)
            end if
            
        end if
        
        if multikey(C_PAUSE) then PauseGame(eng)
    
    loop until multikey(C_ESC)
    
    CleanUp(eng)
    
end sub

sub MainMenu(eng as Engine_t ptr)
    
    dim as integer cB,oB,cX,cY
    dim as string ink
    
    do
        cls
        CPrint("Space Invaders Clone",eng->scr.x/2-80,48,eng)
        
        CPrint("Start Game",eng->scr.x/2-40,80,eng)
        CPrint("Quit Game",eng->scr.x/2-36,112,eng)
        
        if eng->fScr then
            CPrint("Fullscreen: On",eng->scr.x/2-56,160,eng)
        else
            CPrint("Fullscreen: Off",eng->scr.x/2-60,160,eng)
        end if
        
        getmouse cX,cY,,cB
        
        if ((cB and 1) and not(oB and 1)) then
            if Collide(eng->scr.x/2-60,160,120,8,cX,cY,0,0) then
                eng->fScr=-(eng->fScr-1)                
            end if
            
            if Collide(eng->scr.x/2-40,80,80,8,cX,cY,0,0) then
                cls
                exit do
            end if
            
            if Collide(eng->scr.x/2-40,112,72,8,cX,cY,0,0) then
                KillGame(eng)
                end
            end if            
        end if
        oB=cB
        
        screenset eng->cPg, -(eng->cPg-1)
        eng->cPg=(eng->cPg-1)
        
        ink=inkey$+" "
        
        if multikey(C_ESC) then 
            KillGame(eng)
            end
        end if
        
        if ink[0]=255 and ink[1]=107 then
            KillGame(eng)
            end
        end if
        
    loop
    
    cls
    SetScreen(eng)
    
'                1        91        91        91        9
    eng->tm=timer
    cY=eng->scr.y
    do
    
        cls
        
        CPrint ("IT WAS THE END OF THE GIBDIB DYNASTY AND",eng->scr.x/2-160,cY,eng)
        CPrint ("THE WAAKA PEOPLES NEEDED SOMEWHERE TO GO",eng->scr.x/2-160,cY+12,eng)
        
        CPrint ("IT WAS NOT LONG BEFORE THEY FOUND EARTH",eng->scr.x/2-156,cY+36,eng)
        CPrint ("AND THEY SET OUT TO POPULATE IT DESPITE",eng->scr.x/2-156,cY+48,eng)
        CPrint ("THE FACT IT WAS ALREADY POPULATED",eng->scr.x/2-132,cY+60,eng)
        
        CPrint ("YOU MUST STOP THEM AT ANY EXPENCE",eng->scr.x/2-132,cY+84,eng)
        CPrint ("YOU ARE EARTH'S LAST AND FINAL HOPE",eng->scr.x/2-140,cY+96,eng)
        
        if timer-eng->tm > .1 then
            cY-=10*(timer-eng->tm)
            eng->tm=timer
        end if

        screenset eng->cPg, -(eng->cPg-1)
        eng->cPg=(eng->cPg-1) 
            
        if multikey(C_ESC) then exit do
        
    loop while (cY > -104)
    
end sub

dim as Engine_t eng
dim as integer bla,tC
InitGame(@eng)
do
    MainMenu(@eng)
    
    eng.level=0
    do
        tC=0
        for bla=0 to 127
            if multikey(bla) then tC+=1
        next
        if tC=0 then exit do
    loop
    Main(@eng)
    do
        tC=0
        for bla=0 to 127
            if multikey(bla) then tC+=1
        next
        if tC=0 then exit do
    loop
        
loop 
